<?php

#  Copyright (C) 2008  Jason Woofenden
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

require_once('code/tasks.php');

$GLOBALS['tasks_form_recipient'] = "fixme@example.com";



require_once('code/wfpl/template.php');
require_once('code/wfpl/format.php');
require_once('code/wfpl/messages.php');
require_once('code/wfpl/email.php');
require_once('code/db_connect.php');

function description_has_fixmes($description) {
	return (strpos($description, 'FIXME') !== false);
}

function tasks_get_fields() {
	$title = format_oneline($_REQUEST['title']);
	$url = format_oneline($_REQUEST['url']);
	$description = format_unix($_REQUEST['description']);
	$price = format_decimal($_REQUEST['price']);

	tasks_tem_sets($title, $url, $description, $price);

	return array($title, $url, $description, $price);
}

function tasks_tem_sets($title, $url, $description, $price) {
	tem_set('title', $title);
	tem_set('url', $url);
	tem_set('description', $description);
	tem_set('price', $price);
}

function tasks_main() {
	if(!logged_in()) {
		$GLOBALS['url'] = this_url();
		return 'login';
	}

	if(isset($_REQUEST['tasks_id'])) {
		$ret = tasks_display_main();
		if($ret) {
			return $ret;
		}
		tem_show('display_body');
	} else {
		$ret = tasks_edit_main();
		if($ret) {
			return $ret;
		}
		tem_show('edit_body');
	}

	tem_show('main_body');
}

function tasks_display_main() {
	$task_id = format_int($_REQUEST['tasks_id']);;
	$client_id = logged_in();
	if(logged_in_as_contractor()) {
		$row = db_get_row('tasks', 'title,url,description,state,price,client_id,paid,finished_at,tested_at', 'where id=%i', $task_id);
	} else {
		$row = db_get_row('tasks', 'title,url,description,state,price,client_id,paid,finished_at,tested_at', 'where id=%i && client_id=%i', $task_id, $client_id);
	}
	if($row) {
		list($title, $url, $description, $state, $price, $owner_id, $paid, $finished_at, $tested_at) = $row;
		tem_set('task_id', $task_id);
		tem_set('title', $title);
		tem_set('url', $url);
		tem_set('description', $description);
		tem_set('state', task_state_pretty($state));
		tem_set('price', $price);
		if($finished_at) {
			tem_set('finished_at', $finished_at);
			tem_show('finished_at_section');
		}
		if($tested_at) {
			tem_set('tested_at', $tested_at);
			tem_show('tested_at_section');
		}
		if($state == TASK_BUG) {
			tem_show('bug_title');
		} else {
			tem_show('normal_title');
		}
		if(logged_in_as_contractor()) {
			switch($state) {
				case TASK_DRAFT:
				case TASK_NEEDS_CLARIFICATION:
				case TASK_NEEDS_QUOTE:
				case TASK_BUG:
					tem_show('normal_edit_link');
				break;
				case TASK_NEEDS_GO_AHEAD:
					tem_show('approve_price_link');
					tem_show('normal_edit_link');
					tem_show('price_row');
				break;
				case TASK_QUEUED:
					tem_show('normal_edit_link');
					tem_show('working_link');
					tem_show('price_row');
				break;
				case TASK_WORKING:
					tem_show('price_row');
					tem_show('needs_testing_link');
				break;
				case TASK_NEEDS_TESTING:
					if($owner_id == logged_in()) {
						tem_show('finished_link');
					}
					# FALL THROUGH
				case TASK_FINISHED:
					if($paid) {
						tem_show('marked_paid');
					} else {
						tem_show('mark_paid_link');
					}
					tem_show('price_row');
				break;
			}
		} else {
			switch($state) {
				case TASK_DRAFT:
				case TASK_NEEDS_CLARIFICATION:
				case TASK_NEEDS_QUOTE:
				case TASK_BUG:
					tem_show('normal_edit_link');
				break;
				case TASK_NEEDS_GO_AHEAD:
					tem_show('price_row');
					tem_show('approve_price_link');
					tem_show('normal_edit_link');
				break;
				case TASK_QUEUED:
					tem_show('price_row');
					tem_show('warning_edit_link');
				break;
				case TASK_WORKING:
					tem_show('price_row');
				break;
				case TASK_NEEDS_TESTING:
					tem_show('price_row');
					tem_show('finished_link');
				break;
				case TASK_FINISHED:
					tem_show('price_row');
				break;
			}
		}
	} else {
		message("Task #$task_id not found");
		return './';
	}
}

define('MAX_PRIORITY', 2000000000);
define('MIN_PRIORITY', 0);
define('MID_PRIORITY', floor((MAX_PRIORITY - MIN_PRIORITY) / 2));

function new_lowest_priority($client_id) {
	$lowest_ord = db_get_value('tasks', 'ord', 'where client_id=%i && state=%i order by ord asc limit 1', $client_id, TASK_QUEUED);
	if($lowest_ord === false) {
		return MID_PRIORITY;
	}
	if($lowest_ord == MIN_PRIORITY) {
		reprioritize_tasks($client_id); # make room
		$lowest_ord = db_get_value('tasks', 'ord', 'where client_id=%i && state=%i order by ord asc limit 1', $client_id, TASK_QUEUED);
	}
	return MIN_PRIORITY + floor(($lowest_ord - MIN_PRIORITY) / 2);
}

# keep everything in the same order, but space them out so there's room to squeeze things in anywhere
function reprioritize_tasks($client_id) {
	$ids = db_get_column('tasks', 'id', 'where client_id=%i && state=%i order by ord desc, id desc', $client_id, TASK_QUEUED);
	$step = floor((MAX_PRIORITY - MIN_PRIORITY) / (count($ids) + 1));
	$cur = MAX_PRIORITY;
	foreach($ids as $id) {
		$cur -= $step;
		db_update('tasks', 'ord', $cur, 'where id=%i', $id);
	}
}


# pass the task id and one of (up,down,top,bottom)
function prioritize_task($id, $change) {
	$row = db_get_row('tasks', 'client_id,ord', 'where id=%i', $id);
	if(!$row) {
		message('Database error #2242');
		return;
	}
	list($client_id, $ord) = $row;
	switch($change) {
		case 'top':
			list($highest_id, $highest_ord) = db_get_row('tasks', 'id,ord', 'where client_id=%i && state=%i order by ord desc limit 1', $client_id, TASK_QUEUED);
			if($highest_id == $id) {
				message('Already highest priority');
				return;
			}

			if($highest_ord == MAX_PRIORITY) {
				reprioritize_tasks($client_id); # make room
				$highest_ord = db_get_value('tasks', 'ord', 'where client_id=%i && state=%i order by ord desc limit 1', $client_id, TASK_QUEUED);
			}

			$new_ord = MAX_PRIORITY - floor((MAX_PRIORITY - $highest_ord) / 2);
			db_update('tasks', 'ord', $new_ord, 'where id=%i', $id);
			return;
		case 'bottom':
			list($lowest_id, $lowest_ord) = db_get_row('tasks', 'id,ord', 'where client_id=%i && state=%i order by ord asc limit 1', $client_id, TASK_QUEUED);
			if($lowest_id == $id) {
				message('Already lowest priority');
				return $lowest_ord;
			}

			if($lowest_ord == MIN_PRIORITY) {
				reprioritize_tasks($client_id); # make room
				$lowest_ord = db_get_value('tasks', 'ord', 'where client_id=%i && state=%i order by ord asc limit 1', $client_id, TASK_QUEUED);
			}

			$new_ord = MIN_PRIORITY + floor(($lowest_ord - MIN_PRIORITY) / 2);
			db_update('tasks', 'ord', $new_ord, 'where id=%i', $id);
			return;
		case 'up':
		case 'down':
			if($change == 'up') {
				$rows = db_get_rows('tasks', 'id,ord', 'where client_id=%i && state=%i order by ord desc, id desc', $client_id, TASK_QUEUED);
				if($rows[0][0] == $id) {
					message('Already highest priority');
					return;
				}
				if($rows[1][0] == $id) {
					prioritize_task($id, 'top');
					return;
				}
			} else {
				$rows = db_get_rows('tasks', 'id,ord', 'where client_id=%i && state=%i order by ord asc, id asc', $client_id, TASK_QUEUED);
				if($rows[0][0] == $id) {
					message('Already lowest priority');
					return;
				}
				if($rows[1][0] == $id) {
					prioritize_task($id, 'bottom');
					return;
				}
			}
			# find the one we're moving
			$cur_index = 0;
			$done = count($rows);
			for($i = 2; $i < $done ; ++$i) {
				if($rows[$i][0] == $id) {
					$cur_index = $i;
					break;
				}
			}
			$before_ord = $rows[$cur_index - 1][1];
			$before_before_ord = $rows[$cur_index - 2][1];
			if(abs($before_before_ord - $before_ord) < 2) {
				reprioritize_tasks($client_id);
				$before_ord = db_get_value('tasks', 'ord', 'where id=%i', $rows[$cur_index - 1][0]);
				$before_before_ord = db_get_value('tasks', 'ord', 'where id=%i', $rows[$cur_index - 2][0]);
				if($before_before_ord == $before_ord) {
					message('Programmer error #8592');
					return;
				}
			}
			$new_ord = $before_ord + floor(($before_before_ord - $before_ord) / 2);
			db_update('tasks', 'ord', $new_ord, 'where id=%i', $id);
			return;
		default:
			message('invalid change');
			return;
	}

}

function tasks_edit_main() {
	$state = TASK_DRAFT; # will be overwritten
	$client_id = logged_in(); # fixed shortly if we're contractor
	$edit_id = format_int($_REQUEST['tasks_edit_id']);
	unset($_REQUEST['tasks_edit_id']);
	if($edit_id) {
		$owner = db_get_value('tasks', 'client_id', 'where id=%i', $edit_id);
		if(logged_in_as_contractor()) {
			$client_id = $owner;
		} elseif($owner != $client_id) {
			message('Sorry, that task was entered by/for another client.');
			return './';
		}

		# add hidden field for database id of row we're editing
		tem_set('tasks_edit_id', $edit_id);
		tem_show('editing');

		$state = db_get_value('tasks', 'state', 'where id=%i', $edit_id);
	}

	if(isset($_REQUEST['bump'])) {
		switch($_REQUEST['bump']) {
			case 'up':
			case 'down':
			case 'top':
			case 'bottom':
				prioritize_task($edit_id, $_REQUEST['bump']);
				return './';
		}
	}

	if(isset($_REQUEST['tasks_new_bug'])) {
		$state = TASK_BUG;
	}

	if(isset($_REQUEST['tasks_mark_paid_id'])) {
		if(!logged_in_as_contractor()) {
			message("Error: only Jason can mark tasks as paid.");
			return './';
		}
		$id = $_REQUEST['tasks_mark_paid_id'];
		db_update('tasks', 'paid', 1, 'where id=%i', $id);
		message('Marked as paid.');
		return './';
	}

	if(isset($_REQUEST['tasks_approve_price_id'])) {
		$id = $_REQUEST['tasks_approve_price_id'];
		$owner = db_get_value('tasks', 'client_id', 'where id=%i', $id);
		if(logged_in() != $owner) {
			message("Error: can't approve a task entered by/for another client.");
			return './';
		}
		$ord = new_lowest_priority($owner);
		db_update('tasks', 'state,ord', TASK_QUEUED, $ord, 'where id=%i', $id);
		message('Price approved.');
		return './';
	}

	if(isset($_REQUEST['tasks_working_id'])) {
		$id = $_REQUEST['tasks_working_id'];
		if(!logged_in_as_contractor()) {
			message("Error: only Jason can say what he's working on.");
			return './';
		}
		db_update('tasks', 'state', TASK_WORKING, 'where id=%i', $id);
		message('Task marked as "in progress".');
		return './tasks?tasks_id=' . $id;
	}

	if(isset($_REQUEST['tasks_needs_testing_id'])) {
		$id = $_REQUEST['tasks_needs_testing_id'];
		if(!logged_in_as_contractor()) {
			message("Error: only Jason can say when he's done.");
			return './';
		}
		db_update('tasks', 'state,finished_at', TASK_NEEDS_TESTING, date('Y-m-d'), 'where id=%i', $id);
		message('Task awaits testing.');
		return './';
	}

	if(isset($_REQUEST['tasks_finished_id'])) {
		$id = $_REQUEST['tasks_finished_id'];
		$owner = db_get_value('tasks', 'client_id', 'where id=%i', $id);;
		if(logged_in() != $owner) {
			message("Error: can't test a task entered by/for another client.");
			return './';
		}
		db_update('tasks', 'state,tested_at', TASK_FINISHED, date('Y-m-d'), 'where id=%i', $id);
		message('Task marked as finished.');
		# FIXME also mark it as paid if client's balance can cover it
		return './';
	}

	$delete_id = format_int($_REQUEST['tasks_delete_id']);
	unset($_REQUEST['tasks_delete_id']);
	if($delete_id) {
		db_delete('tasks', 'where id=%i', $delete_id);
		message('Task deleted.');

		return './tasks.html';
	}

	if(isset($_REQUEST['title'])) {
		list($title, $url, $description, $price) = tasks_get_fields();
		$queuing = false;

		# FIXME
		if(isset($_REQUEST['save_draft'])) {
			$state = TASK_DRAFT;
		} elseif(isset($_REQUEST['save_bug'])) {
			$state = TASK_BUG;
		} elseif(isset($_REQUEST['save_price_no_tiny']) && logged_in_as_contractor()) {
			$state = TASK_NEEDS_GO_AHEAD;
		} elseif(isset($_REQUEST['save_price']) && logged_in_as_contractor()) {
			$tiny_agreement = db_get_value('people', 'tiny_agreement', 'where id=%i', $client_id);
			if($price <= $tiny_agreement) {
				$state = TASK_QUEUED;
				$queuing = true;
			} else {
				$state = TASK_NEEDS_GO_AHEAD;
			}
		} elseif(isset($_REQUEST['needs_clarification'])) {
			$state = TASK_NEEDS_CLARIFICATION;
		} else { # better be "request_price"
			if(description_has_fixmes($description)) {
				$state = TASK_NEEDS_CLARIFICATION;
				message('The description is not ready to be priced yet because it still contains at least one "FIXME".');
			} else {
				$state = TASK_NEEDS_QUOTE;
			}
		}

		if("you're happy with the POSTed values") { # if you change this change the one above
			if($edit_id) {
				$tables = 'title,url,description,state';
				$values = array($title, $url, $description, $state);
				if(isset($_REQUEST['price']) && logged_in_as_contractor()) {
					$tables .= ',price';
					array_push($values, $price);
				}
				if($queuing) {
					$client_id = db_get_value('tasks', 'client_id', 'where id=%i', $edit_id);
					$tables .= ',ord';
					array_push($values, new_lowest_priority($client_id));
				}
				db_update('tasks', $tables, $values, 'where id=%i', $edit_id);
				message('Task updated.');
			} else {
				# new task
				$paid = 0;
				$client_id = logged_in();
				if(logged_in_as_contractor() && $_REQUEST['client_id']) {
					$client_id = format_int($_REQUEST['client_id']);
				} else {
					# if client entered the task, no price is set
					$price = 0;
				}
				if($state == TASK_QUEUED) {
					$ord = new_lowest_priority($client_id);
				} else {
					$ord = 0;
				}
				db_insert('tasks', 'client_id,title,url,description,state,paid,price,ord', $client_id, $title, $url, $description, $state, $paid, $price, $ord);
				message('Task saved.');
			}
			if($GLOBALS['tasks_form_recipient'] != "fixme@example.com") {
				$to = $GLOBALS['tasks_form_recipient'];
				$from = $to;
				$reply_to = '';
				if(isset($_REQUEST['email']) and valid_email($_REQUEST['email'])) {
					$reply_to = $_REQUEST['email'];
					if($_REQUEST['name'] and ereg('^[a-zA-Z0-9_\' -]*$', $_REQUEST['name']) !== false) {
						$reply_to = "$_REQUEST[name] <$reply_to>";
					}
				}
				$subject = 'tasks form submitted';
				$message = tem_run('tasks.email.txt');
				$cc = '';
				$bcc = '';
				if(email($from, $to, $subject, $message, $reply_to, $cc, $bcc)) {
					message('Due to an internal error, your message could not be sent. Please try again later.');
					$error = true;
				}
			}
			if($error !== true) {
				return './';
			}
		}
		# otherwise, we display the form again. tasks_get_fields() has
		# already put the posted values back into the template engine, so they will
		# show up in the form fields. You should add some message asking people to
		# fix their entry in whatever way you require.
	} elseif($edit_id) {
		# we've recieved an edit id, but no data. So we grab the values to be edited from the database
		list($title, $url, $description, $state, $paid) = db_get_row('tasks', 'title,url,description,state,price', 'where id=%i', $edit_id);
		tasks_tem_sets($title, $url, $description, $price);
	} else {
		# form not submitted, you can set default values like so:
		#tem_set('client_id', 'Yes');
	}

	# display header
	if($edit_id) {
		tem_show('edit_msg');
	} elseif($state == TASK_BUG) {
		tem_show('bug_msg');
	} else {
		tem_show('new_msg');
		if(logged_in_as_contractor()) {
			pulldown('client_id', db_get_rows('people', 'id,name', 'where id > 1'), PULLDOWN_2D);
			tem_set('client_id', format_int($_REQUEST['client_id']));
			tem_show('client_row');
		}
	}

	# display instructions
	if($state == TASK_BUG) {
		tem_show('bug_instructions');
		if(logged_in_as_contractor()) {
			tem_show('price_field');
			tem_show('contractor_submits');
		} else {
			tem_show('bug_submit');
		}
	} elseif($state == TASK_NEEDS_QUOTE && logged_in_as_contractor()) {
		tem_show('set_price_instructions');
		tem_show('price_field');
		tem_show('contractor_submits');
	} else {
		if(description_has_fixmes($description)) {
			tem_show('fixme_instructions');
		} else {
			tem_show('normal_instructions');
		}
		if(logged_in_as_contractor()) {
			tem_show('contractor_submits');
			switch($state) {
				case TASK_DRAFT:
				case TASK_NEEDS_CLARIFICATION:
				case TASK_NEEDS_QUOTE:
				case TASK_NEEDS_GO_AHEAD:
				case TASK_QUEUED:
				case TASK_BUG:
				tem_show('price_field');
			}
		} else {
			tem_show('normal_submits');
		}
	}
}

?>
